;* Title:		Square Wave Generator
;* Version:		2.0 Beta
;* Target:		AT90S2313
;* CLK frequency:	10.000MHz
;* Authors:		Leonid Ridiko (wubblick@yahoo.com)
;			Andre Birua   (birua@hotmail.com)

.include "2313def.inc"

;***** Constantes

;General:

.equ	MinD	=15		;minimum D, x0.1uS
.equ	MinDm	=1		;minimum D in meander mode, x0.1uS
.equ	MinP	=30		;minimum P, x0.1uS
.equ	MinE	=30		;minimum E, x0.1uS
.equ	DelOff	=1		;delay offset, x0.1uS
.equ	MinN	=1		;minimum n
.equ	MaxN	=255		;maximum n
.equ	MaxNS	=50		;maximum n step
.equ	MaxDPE	=9999999	;maximum D,P,E x0.1uS
.equ	MaxS	=500000		;maximum step, x0.1uS

.equ	MaxW	=50000		;max count in word format
.equ	MinDe	=100		;min delta = OCR1A - TCNT1 for reload
.equ	T0Val	= -98		;x1024x0.0001=10mS program timer
.equ	ARCnV	=16		;slow autorepeat count
.equ	ARDel	=80		;initial autorepeat delay,	x10mS
.equ	ARSlw	=18		;slow autorepeat rate,		x10mS
.equ	ARFst	=6		;fast autorepeat rate,		x10mS
.equ	DebV	=3		;debounce delay,		x10mS

;Keyboard scancodes:

.equ	K_NO	=0x00		;no press
.equ	K_EX	=0x01		;key EXIT code
.equ	K_DN	=0x02		;key DOWN code
.equ	K_UP	=0x04		;key UP code
.equ	K_EN	=0x08		;key ENTER code

;Menu structure:

;[d  0]	<-> [dS 6]
;  |
;[P  1] <-> [PS 7]
;  |
;[SH 2] <-> [EEP 8]
;  |
;[SY 3] <-> [Go 9]
;  |
;[E  4] <-> [ES 10]
;  |
;[n  5] <-> [nS 11]

.equ	MnuD	=0		;edit Duration
.equ	MnuP	=1		;edit Period
.equ	MnuSH	=2		;edit SHape
.equ	MnuSY	=3		;edit SYnc. mode
.equ	MnuE	=4		;edit dElay
.equ	MnuN	=5		;edit n
.equ	MnuDS	=6		;edit Duration Step
.equ	MnuPS	=7		;edit Period Step
.equ	MnuEE	=8		;store values in EEProm
.equ	MnuGo	=9		;Go series
.equ	MnuES	=10		;edit dElay Step
.equ	MnuNS	=11		;edit n Step

.equ	IniMnu	=MnuD		;initial menu item

;ValSH values:

.equ	ShMe	=1		;meander mode
.equ	ShHi	=2		;hi pulse
.equ	ShLo	=3		;lo pulse
.equ	ShFo	=4		;lo level
.equ	ShFi	=5		;hi level

;ValSY values:

.equ	SyAu	=1		;sync. auto
.equ	SyHi	=2		;ext. sync. lo-hi
.equ	SyLo	=3		;ext. sync. hi-lo

;Max menu items:

.equ	MaxSH	=5		;max SH menu item
.equ	MaxSY	=3		;max SY menu item
.equ	MaxEE	=1		;max EEP and Go menu item

;Periphery control codes:

.equ	CapH	=0x41		;capture L-H, CTC1 = 0 TCCR1B value
.equ	CapL	=0x01		;capture H-L, CTC1 = 0 TCCR1B value
.equ	Comp	=0x09		;compare,     CTC1 = 1 TCCR1B value
.equ	IntDs	=0x00		;int disable           TIMSK  value
.equ	IntCp	=0x08		;capture int enable    TIMSK  value
.equ	IntCm	=0x40		;compare int enable    TIMSK  value
.equ	OutTg	=0x40		;out toggle            TCCR1A value
.equ	OutHi	=0xC0		;out high              TCCR1A value
.equ	OutLo	=0x80		;out low               TCCR1A value
	
;***** Variables

.DSEG	;data segment (internal SRAM)

RAMBEG:				;internal SRAM: RAMBEG..RAMEND

Vals:				;values array	[d,P,SH,SY,E,n]

ValD:	.byte 3			;duration
ValP:	.byte 3			;period
ValSH:	.byte 3			;shape
ValSY:	.byte 3			;synchronization
ValE:	.byte 3			;delay
ValN:	.byte 3			;n

Stps:	.byte ValN-ValD+3	;steps array

.equ	Soff	=(Stps-Vals)/3	;steps array index offset

Mem0:	.byte 4			;Memory copy of TH0,TL0,TN0,FLAGS0
Mem1:	.byte 4			;Memory copy of TH1,TL1,TN1,FLAGS1

NS:	.byte 1 		;* n series load value

Buff:	.byte 3			;edit buffer
Step:	.byte 3			;plus/minus step
Max:	.byte 3			;max limit
Min:	.byte 3			;min limit

Dig:	.byte 10		;LCD digits copy

KBD:	.byte 5			;keyboard data structure
.equ	Lc	=0		;LastCode offset
.equ	Tc	=1		;TempCode offset
.equ	ARCnt	=2		;autorepeat counter offset
.equ	DebTM	=3		;debounce timer offset
.equ	KeyTM	=4		;key timer offset

.ESEG	;EEPROM initial values

Enone:	.DB 0			;address 0 not used
EVals:				;values array

EValD:	.DB 100,0,0		;duration
EValP:	.DB 200,0,0		;period
EValSH:	.DB 2,0,0		;shape
EValSY:	.DB 1,0,0		;synchronization
EValE:	.DB 100,0,0		;delay
EValN:	.DB 1,0,0		;n

EStps:				;steps array

EStpD:	.DB 1,0,0		;duration step
EStpP:	.DB 1,0,0		;period step
EStpSH:	.DB 1,0,0		;store menu index
EStpSY:	.DB 0,0,0		;not used
EStpE:	.DB 1,0,0		;delay step
EStpN:	.DB 1,0,0		;n step
Eend:				;end of used EEPROM

;***** Global Register Variables

;r0 used with lpm instruction
		
.def	TL0	=r01		;* timer load value L (0)
.def	TH0	=r02		;* timer load value H (0)
.def	TN0	=r03		;* cycles number (0)
.def	TL1	=r04		;* timer load value L (1)
.def	TH1	=r05		;* timer load value H (1)
.def	TN1	=r06		;* cycles number (1)
.def	PCNT	=r07		;* program counter
.def	RgCAP	=r08		;* capture mode restore code
.def	NSER	=r09		;* n series counter

.def	tempQ	=r10		;#
.def	tempP	=r11		;#
.def	XSREG	=r12		;#* SREG save register

.def	tempA	=r13
.def	tempB	=r14
.def	tempC	=r15

.def	tempR	=r16		;#*
.def	temp	=r17		
.def	tempL	=r18		
.def	tempM	=r19		
.def	tempH	=r20		
.def	tempD	=r21
.def	tempE	=r22
.def	tempF	=r23

.def	Mode	=r24
.equ	Loa	=0		;low active level flag
.equ	Fix	=1		;fixed output level flag
.equ	Mea	=2		;meander flag
.equ	Exo	=3		;external sync. fall activate flag
.equ	Exs	=4		;external sync. flag

.def	FLAGS0	=r25		;* flags 0	
.def	FLAGS1	=r26		;* flags 1	
.equ	Sh	=0		;short mode flag
.equ	ShEx	=1		;short ext. mode flag
.equ	Ex	=2		;ext. mode flag
.equ	Lev	=3		;low active level flag

.def	FLAGS2	=r27		;flags 2	
.equ	NEWPR	=0		;new press flag
.equ	UPDD	=1		;display update flag
.equ	EDT	=2		;edited flag
.equ	UPD	=3		;values update flag

.def	Menu	=r29		;menu item

;r28 used as YL register
;r30,r31 used as Z register

;* - used in COMPARE interrupt
;# - used in CAPTURE interrupt

;***** Port Definitions

;DIRx = 1 - out
;PUPx = 1 - IN:pull-up on; OUT: =1

;Port B:

.equ	DIRB	=0b00001111	;directions setup
.equ	PUPB	=0b11110111	;pull-ups and level setup

.equ	DATA	=PB0		;LCD serial data line
.equ	LOAD	=PB1		;LCD load line
.equ	CLK	=PB2		;LCD clock line
.equ	OUT	=PINB3		;generator out
.equ	RETL	=PINB4		;keyboard return line

;Port D:

.equ	DIRD	=0b00100000	;directions setup
.equ	PUPD	=0b11111111	;pull-ups and level setup
.equ	SND	=PD5		;sound out

;***** Macros

.macro	clbr	;clear bit in register
	cbr @0,exp2(@1)
.endm

.macro	stbr	;set bit in register
	sbr @0,exp2(@1)
.endm

.macro	bbrc	;branch if bit in register clear
	sbrs @0,@1
	rjmp @2
.endm

.macro	bbrs	;branch if bit in register set
	sbrc @0,@1
	rjmp @2
.endm

.macro	bbic	;branch if bit in I/O clear
	sbis @0,@1
	rjmp @2
.endm

.macro	bbis	;branch if bit in I/O set
	sbic @0,@1
	rjmp @2
.endm

.macro	ldmi	;load tempH,M,L with 24-bit data
	ldi tempL,low(@0)	
	ldi tempM,high(@0)	
	ldi tempH,byte3(@0)	
.endm

.macro	ldei	;load tempF,E,D with 24-bit data
	ldi tempD,low(@0)	
	ldi tempE,high(@0)	
	ldi tempF,byte3(@0)	
.endm

.macro	IntDis
;timers interrupts disable
	ldi	temp,IntDs
	out	TIMSK,temp
.endm

.macro	IntEn
;compare interrupt enable for meander mode
	ldi	temp,IntCm
	out	TIFR,temp	;clear pending interrupt
	out	TIMSK,temp	;compare interrupt enable
.endm

.macro	IntCAP
;* capture interrupt enable
	out	TCCR1B,RgCAP	;clear CTC1
	ldi	tempR,IntCp
	out	TIFR,tempR	;clear pending interrupt
	out	TIMSK,tempR	;capture interrupt enable
.endm

.macro	IntCOM
;# compare interrupt enable
	ldi	tempR,Comp
	out	TCCR1B,tempR	;set CTC1
	ldi	tempR,IntCm
	out	TIFR,tempR	;clear pending interrupt
	out	TIMSK,tempR	;compare interrupt enable
.endm

.macro	OutToggle
;#* OC1 toggle	
	ldi	tempR,OutTg
	out	TCCR1A,tempR
.endm	
	
.macro	OutH	
;#* OC1 -> H	
	ldi	tempR,OutHi
	out	TCCR1A,tempR
.endm	
	
.macro	OutL
;#* OC1 -> L	
	ldi	tempR,OutLo
	out	TCCR1A,tempR
.endm

;***** Interrupt Vectors

.CSEG
.org	0			;Reset vector
	rjmp	INIT		
.org	ICP1addr		;Capture interrupt vector
	rjmp	CPTURE		
.org	OC1addr			;Compare interrupt vector

;***** Interrupts Service Routines

;***** Compare interrupt handle (starting at vector):

CMPARE:	sbic	PINB,OUT	;check up out pin
	rjmp	Phase1
	
;Short mode (0):	
	
Phase0:	out	OCR1AH,TH0	;compare register reload (0)
	out	OCR1AL,TL0
	sbrc	FLAGS0,Sh	;check Sh flag
	reti

;Ext. sync. mode (0):

PhEx0:	in	XSREG,SREG	;save SREG
	sbrs	FLAGS0,ShEx	;check ShEx flag
	rjmp	PhLg0
Exdo0:	tst	NSER		;check NSER
	brne	Exn0
	lds	NSER,NS		;NSER load
Exn0:	dec	NSER
	breq	Exz0
	bbrc	FLAGS0,ShEx,Pr0
	rjmp	Ret0	
Exz0:	OutL			;OC1 clear on compare
	IntCAP			;capture interrupt enable
Ret0:	out	SREG,XSREG	;restore SREG
	reti
	
;Long mode (0):

PhLg0:	tst	PCNT		;test program counter
	brne	Pcnt0
	sbrc	FLAGS0,Ex	;check Ex flag
	rjmp	Exdo0
	
Pr0:	mov	PCNT,TN0	;program counter reload
	rjmp	Pcld0
Pcnt0:	dec	PCNT
Pcld0:	tst	PCNT
	breq	Pcze0
	OutL			;OC1 clear on compare
	rjmp	Pe01
	
Pcze0:	OutToggle		;OC1 toggle on compare
	sbrs	FLAGS0,Ex	;check Ex flag
	rjmp	Ret0
	
Pc01:	tst	NSER
	brne	Ret0
	
Pe01:	ldi	tempR,high(MaxW) ;load compare register with const.
	out	OCR1AH,tempR	
	ldi	tempR, low(MaxW)
	out	OCR1AL,tempR
	rjmp	Ret0
	
;Short mode (1):	
	
Phase1:	out	OCR1AH,TH1	;compare register reload (1)
	out	OCR1AL,TL1
	sbrc	FLAGS1,Sh	;check Sh flag
	reti

;Ext. sync. mode (1):

PhEx1:	in	XSREG,SREG	;save SREG
	sbrs	FLAGS1,ShEx	;check ShEx flag
	rjmp	PhLg1
Exdo1:	tst	NSER		;check NSER
	brne	Exn1
	lds	NSER,NS		;NSER load
Exn1:	dec	NSER
	breq	Exz1
	bbrc	FLAGS1,ShEx,Pr1
	rjmp	Ret1	
Exz1:	OutH			;OC1 set on compare
	IntCAP			;capture interrupt enable
Ret1:	out	SREG,XSREG	;restore SREG
	reti

;Long mode (1):

PhLg1:	tst	PCNT		;test program counter
	brne	Pcnt1
	sbrc	FLAGS1,Ex	;check Ex flag
	rjmp	Exdo1
	
Pr1:	mov	PCNT,TN1	;program counter reload
	rjmp	Pcld1
Pcnt1:	dec	PCNT
Pcld1:	tst	PCNT
	breq	Pcze1
	OutH			;OC1 set on compare
	rjmp	Pe01
	
Pcze1:	OutToggle		;OC1 toggle on compare
	sbrs	FLAGS1,Ex	;check Ex flag
	rjmp	Ret1
	rjmp	Pc01

;***** Capture interrupt handle:	
	
CPTURE:	in	XSREG,SREG	;save SREG
	in	tempQ,ICR1L	;read capture register low  byte
	in	tempP,ICR1H	;read capture register high byte	
	sbrc	FLAGS0,Lev	;check Lev flag (FLAGS0 or FLAGS1)
	rjmp	Cap1
	
;Short delay (0):	
	
Cap0:	add	tempQ,TL0	;add delay
	adc	tempP,TH0
	out	OCR1AH,tempP	;load compare register
	out	OCR1AL,tempQ
	OutToggle		;OC1 toggle on compare
	IntCOM			;compare interrupt enable
	out	SREG,XSREG	;restore SREG
	sbrc	FLAGS0,ShEx	;check ShEx flag
	reti
	
;Long delay (0):	
	
CLg0:	mov	PCNT,TN0	;program counter load
	OutL			;OC1 clear on compare
	out	SREG,XSREG	;restore SREG
	reti
	
;Short delay (1):	

Cap1:	add	tempQ,TL1	;add delay
	adc	tempP,TH1
	out	OCR1AH,tempP	;load compare register
	out	OCR1AL,tempQ
	OutToggle		;OC1 toggle on compare
	IntCOM			;compare interrupt enable
	out	SREG,XSREG	;restore SREG
	sbrc	FLAGS1,ShEx	;check ShEx flag
	reti
	
;Long delay (1):	
	
CLg1:	mov	PCNT,TN1	;program counter load
	OutH			;OC1 set on compare
	out	SREG,XSREG	;restore SREG
	reti
	
;***** Program Execution Starts Here

INIT:	wdr			;Watchdog reset
	ldi	temp,0x0A	;Watchdog enable, 64mS
	out	WDTCR,temp
	
;Ports setup:	
	
	ldi	temp,PUPB
	out	PORTB,temp	;Init PORTB and on/off pullup
	ldi	temp,DIRB	
	out	DDRB,temp	;Set PORTB direction
	
	ldi	temp,PUPD
	out	PORTD,temp	;Init PORTD and on/off pullup
	ldi	temp,DIRD
	out	DDRD,temp	;Set PORTD direction
	
;Variables init:
	
	clr	FLAGS2		;clear flags
	
	ldi	YL,RAMBEG
clear:	st	Y+,FLAGS2	;clear internal SRAM
	cpi	YL,RAMEND
	brne	clear
	
	out	SPL,YL		;locate stack
	
	ldi	temp,K_NO
	ldi	YL,KBD
	std	Y+Lc,temp	;LastCode = no press code
	std	Y+Tc,temp	;TempCode = no press code
	
;Periphery setup:

	ldi	temp,0x05
	out	TCCR0,temp	;CK/1024 for TC0
	ldi	temp,Comp
	out	TCCR1B,temp	;CK/1 for TC1
	
;Display init:

	ldi	temp,0x0F	;temp <- BLK register address
	rcall	LCD_WA		;write address
	ldi	temp,0x0F	;temp <- 0x0F - enable bus
	rcall	LCD_WN		;write nibble
	rcall	DLoad		;display "0000000000"
	
;Read EEPROM:

	ldi	tempE,EVals	;load initial EEPROM array address
	ldi	YL,Vals		;load initial RAM array address
	
ee_rd:	rcall	EE_Rdy		;wait for EEPROM ready and read EEPROM
	st	Y+,tempD	;store byte in RAM
	brne	ee_rd

;Validate all parameters (except D):

	ldi	Menu,MnuNS	;last menu item
valall:	rcall	SetMd		;set mode
	rcall	ValidB		;validate buffer and save
	rcall	MakeMd		;make Mode variable
	dec	Menu
	brne	valall
	
;Current mode setup:

	ldi	Menu,IniMnu	;display initial menu item
	rcall	ChMd		;setup mode, validate D, SetMd
	
	ldi	tempE,35
	rcall	Tone		;initial beep
	sei			;enable global interrupt
	
;***** Main Loop:

MAIN:	rcall	SCAN		;scan keyboard
	ldi	YL,KBD		;keyboard data structure base
	ldd	tempL,Y+Lc	
	cp	tempL,temp	;scancode = LastCode ?
	breq	Hold		;branch if same key
	
	ldd	tempL,Y+Tc
	cp	tempL,temp	;scancode = TempCode ?
	brne	NewP		;branch if new key
	
	ldd	tempL,Y+DebTM
	tst	tempL		;check debounce timer
	brne	Hold
	
	std	Y+Lc,temp	;LastCode <- scancode
	stbr	FLAGS2,NEWPR	;set new press flag
	ldi	tempH,ARDel	;autorepeat delay value
	ldi	tempL,ARCnV
	rjmp	Stac		;go to store ARCnt
	
NewP:	std	Y+Tc,temp	;TempCode <- scancode
	ldi	tempL,DebV
	std	Y+DebTM,tempL	;debounce timer load
	
Hold:	clbr	FLAGS2,NEWPR	;clear new press flag
	ldd	tempL,Y+KeyTM
	tst	tempL		;check key timer
	brne	Proc
	stbr	FLAGS2,NEWPR	;set new press flag
	ldd	tempL,Y+ARCnt
	tst	tempL
	breq	Fast		;fast autorepeat if count is over
	dec	tempL		;dec autorepeat counter
	ldi	tempH,ARSlw	;slow autorepeat rate
	
Stac:	std	Y+ARCnt,tempL	;store autorepeat counter ARCnt
	rjmp	Stkt		;go to store KeyTM
	
Fast:	ldi	tempH,ARFst	;fast autorepeat rate
Stkt:	std	Y+KeyTM,tempH	;store key timer KeyTM
Proc:
	
;KBD+Lc	= true scancode,
;NEWPR	= 1 if new press

;Process key functions:
	
	bbrc	FLAGS2,NEWPR,NoPr
	ldd	temp,Y+Lc	;temp <- LastCode
	cpi	temp,K_NO
	breq	NoPr		;skip if no press
	rcall	Sound		;key beep
	cpi	temp,K_UP
	brne	Pro1
	rcall	Do_UP		;key UP processing
Pro1:	cpi	temp,K_DN
	brne	Pro2
	rcall	Do_DN		;key DN processing
Pro2:	cpi	temp,K_EX
	brne	Pro3
	rcall	Do_EX		;key EX processing
Pro3:	cpi	temp,K_EN
	brne	NoPr
	rcall	Do_EN		;key EN procesing
NoPr:

;Process timers:	
	
	in	temp,TIFR
	bbrc	temp,TOV0,Novf	;check for Timer 0 overflow
	stbr	temp,TOV0	;Timer 0 overflow flag clear
	out	TIFR,temp
		
	ldi	temp,T0Val
	out	TCNT0,temp	;Timer 0 reload
	ldi	YL,KBD		;keyboard data structure base
	
	ldd	temp,Y+DebTM
	tst	temp		;check debounce timer		 
	breq	Tmr1
	dec	temp		;advance debounce timer
	std	Y+DebTM,temp

Tmr1:	ldd	temp,Y+KeyTM
	tst	temp		;check key timer		 
	breq	Tmr2
	dec	temp		;advance key timer
	std	Y+KeyTM,temp
Tmr2:	
Novf:

;Update display:

	bbrc	FLAGS2,UPDD,NoUpd ;check up update flag
	clbr	FLAGS2,UPDD	  ;clear update flag
	rcall	Disp		  ;update display
NoUpd:

;Safe update values: 

	bbrc	FLAGS2,UPD,Skip	;check up values update flag
	mov	temp,TN0
	or	temp,TN1
	breq	Short		;TN0 and TN1 = 0
	
	tst	PCNT
	brne	Skip		;wait for PCNT = 0
	rcall	Delta		;delta = OCR1A - TCNT1
	brcs	Skip		;skip if delta < MinDe
	
Short:	bbrc	Mode,Mea,DoLoad	;not meander mode, immediate load
	ldi	YL,Mem0
	ldd	temp,Y+2	;load new TN0
	tst	temp		;check mode (short or long)		
	breq	smmd		
lmmd:	IntEn			;long meander mode, int. enable
	rjmp	DoLoad
smmd:	IntDis			;short meander mode, int disable
	ldd	tempH,Y+0
	ldd	tempL,Y+1
	out	OCR1AH,tempH	;load capture register in short mode
	out	OCR1AL,tempL

DoLoad:	rcall	ImLoad		;copy data from memory buffer
	clbr	FLAGS2,UPD	;all updated, clear flag
Skip:
	
;Process series:

	cpi	Menu,MnuGo	;check for Go menu
	brne	NoGo
	tst	NSER		;check NSER
	brne	NoGo
	ldi	Menu,MnuSY
	rjmp	RetMnu		;return to SY menu
NoGo:	

;Save values:

	cpi	Menu,MnuEE	;check for EEP menu
	brne	NoSt	

;Write to EEPROM:
	
	ldi	tempE,EVals	;load initial EEPROM array address
	ldi	YL,Vals		;load initial RAM array address
	
ee_wr:	rcall	EE_Rdy		;wait for EEPROM ready and read EEPROM
	ld	temp,Y+		;read RAM
	cpse	tempD,temp
	rjmp	do_wr		;write if different
	rjmp	sk_wr		;skip write if same
do_wr:	out	EEDR,temp	;load data
	cli			;interrupts disable
	sbi 	EECR,EEMWE	;master write enable
	sbi	EECR,EEWE	;strobe
	sei			;interrupts enable
sk_wr:	brne	ee_wr

	ldi	Menu,MnuSH
	ldi	tempE,50	;data save melody
	rcall	Tone
	ldi	tempE,45
	rcall	Tone
RetMnu:	ldi	tempE,35
	rcall	Tone
	rcall	SetMd		;return back
NoSt:

;Watchdog restart:

	wdr			;watchdog restart
	rjmp	MAIN
	
;***** Subroutines

;Key "UP" processing:

Do_UP:	cpi	Menu,MnuDS	;step edit ?
	brlo	upns
	
;Step change (1 -> 2 -> 5 -> 10 -> 20..)

	rcall	LdBuff		;tempH,M,L <- Buff
	lsl	tempL
	rol	tempM
	rol	tempH		;Buff * 2
	rcall	Search
	cpi	temp,2
	brne	vbuf		;validate buffer
	ldi	YL,Buff
	rcall	LdABC		;tempC,B,A <- Buff	
	lsr	tempC
	ror	tempB
	ror	tempA		;Buff / 2
	add	tempL,tempA
	adc	tempM,tempB
	adc	tempH,tempC	;Buff * 2.5
	rjmp	vbuf		;validate buffer

;Buff + Step, align value to step

upns:	rcall	Rem		;tempH,M,L <- Buff, tempF,E,D <- Step, tempC,B,A <- Rem
	sub	tempL,tempA	;Buff - Rem(Buff/Step)
	sbc	tempM,tempB
	sbc	tempH,tempC
	add	tempL,tempD	;Buff + Step
	adc	tempM,tempE
	adc	tempH,tempF

vbuf:	rcall	Valid		;validate tempH,M,L
	rcall	Assume		;Assume new parameters
	stbr	FLAGS2,UPDD	;set display update flag
	stbr	FLAGS2,EDT	;set edited flag
EndOP:	ldi	temp,K_NO	;end of processing
	ret
	
;Key "DN" processing:
		
Do_DN:	cpi	Menu,MnuDS	;step edit ?
	brlo	dnns
	
;Step change (..20 -> 10 -> 5 -> 2 -> 1)

	rcall	LdBuff		;load Buff to tempH,M,L
	lsr	tempH
	ror	tempM
	ror	tempL		;Buff / 2
	rcall	Search
	cpi	temp,5
	brne	vbuf		;validate buffer
	rcall	LdBuff		;load Buff to tempH,M,L	
	lsl	tempL
	rol	tempM
	rol	tempH		;Buff * 2
	ldi	tempD,5
	clr	tempE
	clr	tempF
	rcall	Div24
	com	tempL
	com	tempM
	com	tempH		;Buff / 2.5
	rjmp	vbuf		;validate buffer

;Buff - Step, align value to step

dnns:	rcall	Rem		;tempH,M,L <- Buff, tempF,E,D <- Step, tempC,B,A <- Rem
	mov	temp,tempA
	or	temp,tempB
	or	temp,tempC
	breq	dn0
	sub	tempL,tempA	;Buff - Rem(Buff/Step)
	sbc	tempM,tempB
	sbc	tempH,tempC
	rjmp	dn1		;validate buffer	
dn0:	sub	tempL,tempD	;Buff - Step
	sbc	tempM,tempE
	sbc	tempH,tempF
dn1:
	
;Auto change step
	
	cpi	tempD,1		;tempD = 1 means Step = 1
	breq	vbuf		;Step = 1, nothing to change
	rcall	ChkMin		;check for Min
	brcc	vbuf		;no overflow, validate buffer
	push	Menu		;save curent menu item
	subi	Menu,-Soff	;change step mode
	rcall	SetMd		;set change step mode
	rcall	Disp		;fill [Dig...]
	rcall	ErrB		;error bell
	rcall	Do_DN		;step down
	pop	Menu
	rjmp	SetMd		;restore mode
	
;tempH,L,M = Buff, tempF,E,D = Step, tempC,B,A = rem(Buff/Step)

Rem:	rcall	LdBuff		;load Buff to tempH,M,L
	ldi	YL,Step
	rcall	LdDEF
	rcall	div24
	
;Load tempH,M,L from Buff

LdBuff:	ldi	YL,Buff

;Load tempH,M,L from [Y+2],[Y+1],[Y+0]

LdLMN:	ld	tempL,Y+
	ld	tempM,Y+
	ld	tempH,Y+
	ret
	
;Load tempC,B,A from [Y+2],[Y+1],[Y+0]

LdABC:	ld	tempA,Y+
	ld	tempB,Y+
	ld	tempC,Y+
	ret
	
;Load tempF,E,D from [Y+2],[Y+1],[Y+0]

LdDEF:	ld	tempD,Y+
	ld	tempE,Y+
	ld	tempF,Y+
divret:	ret
		
;Divide tempH,M,L / tempF,E,D = ~tempH,M,L , tempC,B,A

div24:	clr	tempA		;clear remainder Low byte
	clr	tempB
	sub	tempC,tempC	;clear remainder High byte and carry
	
	ldi	temp,25		;init loop counter
	
div1:	rol	tempL		;shift left dividend
	rol	tempM
	rol	tempH
	dec	temp		;decrement counter
	breq	divret
	rol	tempA		;shift dividend into remainder
	rol	tempB
	rol	tempC
	sub	tempA,tempD	;remainder = remainder - divisor
	sbc	tempB,tempE	
	sbc	tempC,tempF	
	brcc	div1		;if result negative
	add	tempA,tempD	; restore remainder
	adc	tempB,tempE
	adc	tempC,tempF
	rjmp	div1

;Search for first non-zero digit:

Search:	ldi	YL,Dig+3
srch:	ld	temp,Y+
	andi	temp,~Pt
	breq	srch
	cpi	temp,BLANK
	breq	srch
	ret
	
;Validate:
;Input: [Buff], [Max], [Min] - limits
;Out:   [Buff], [Vals[Menu]]

ValidB:	rcall	LdBuff		;tempH,M,L <- [Buff]

;Validate:
;Input: tempH,M,L, [Max], [Min] - limits
;Out:   [Buff], [Vals[Menu]]

Valid:	rcall	ChkMin		;check for Min
	brcs	RstBf
	rcall	ChkMax		;check for Max
	brcc	ValOk	
	
RstBf:	rcall	LdBuff		;restore buffer	
	rcall	ErrB		;error bell
	rcall	ChkMin		;check for Min
	brcs	LimBf
	rcall	ChkMax		;check for Max
	brcc	ValOk	
	
LimBf:	mov	tempL,tempA	;limit buffer
	mov	tempM,tempB
	mov	tempH,tempC
	rcall	ErrB		;error bell

ValOk:	ldi	YL,Buff
	rcall	StY		;save buffer
	ldi	YL,Vals
	add	YL,Menu
	add	YL,Menu
	add	YL,Menu		;YL = Vals + Menu * 3
	rjmp	StY		;Save Buff (tempH,M,L) to [Vals+temp*3]
	
;Compare tempH,M,L and Min:
;Out: C = 1 if limit exceeded
	
ChkMin:	ldi	YL,Min
	rcall	LdABC		;tempC,B,A <- Min
	ldi	temp,0xC0	;0xC00000 is max negative
	cp	temp,tempH
	brcs	chmr		
	cp	tempL,tempA
	cpc	tempM,tempB
	cpc	tempH,tempC
chmr:	ret

;Compare tempH,M,L and Max:
;Out: C = 1 if limit exceeded

ChkMax:	ldi	YL,Max
	rcall	LdABC		;tempC,B,A <- Max
	cp	tempA,tempL
	cpc	tempB,tempM
	cpc	tempC,tempH
	ret

;Make Mode variable
;Input: ValSH, ValSY
;Out:   Mode

;Mode format:
;
;ValSH:       ShMe  ShHi  ShLo  ShFo  ShFi
;
;Mode.0	(Loa)  0     0     1     0     1	;low passive level flag
;Mode.1	(Fix)  0     0     0     1     1	;fixed output level flag
;Mode.2	(Mea)  1     0     0     0     0	;meander mode flag
;
;ValSY:       SyAu  SyHi  SyLo
;
;Mode.3	(Exo)  1     0     1			;fall act. sync. flag
;Mode.4	(Exs)  0     1     1			;ext. sync. flag
	
MakeMd:	ldi	YL,ValSH
	rcall	ldLMN		;tempL <- ValSH
	subi	tempL,2
	brcc	nme
	andi	tempL,4
nme:	mov	Mode,tempL	;setup Loa, Fix, Mea
	ldi	YL,ValSY
	rcall	ldLMN		;tempL <- ValSY
	swap	tempL
	lsr	tempL
	sbrs	Mode,Mea	;Exo, Exs = 0 in meander mode
	or	Mode,tempL	;setup Exo, Exs
	ret

;Change mode (SH or SY):

ChMd:	IntDis			;interrupts disable
	clr	PCNT		;clear program counter
	clr	NSER		;clear n series counter
	rcall	MakeMd		;make Mode variable
	
;Validate D	
	
	push	Menu
	ldi	Menu,MnuD
	rcall	SetMd		;Buff <- D
	rcall	ValidB		;validate D, save D
	pop	Menu
	rcall	SetMd		;restore mode
	
;Setup output level
	
	bbrc	Mode,Loa,hoa	;check up level flag
	OutH			;OC1 high
	rjmp	fxp
hoa:	OutL			;OC1 low
fxp:	sbrc	Mode,Fix
	ret			;return if fixed output level
	bbrs	Mode,Mea,syna	;meander mode, auto sync.
	bbrc	Mode,Exs,syna	;Exs=0, auto sync.
	
;External sync.	
	
	bbrc	Mode,Exo,exi	;check ext. sync. polarity
	ldi	temp,CapL	;capture H-L
	mov	RgCAP,temp	;save capture setup value
	rjmp	syne
exi:	ldi	temp,CapH	;capture L-H
	mov	RgCAP,temp	;save capture setup value
syne:	rcall	LoadD		;load D to memory buffer
	rcall	LoadE		;load E to memory buffer
	rcall	LoadN		;load N to NS
	rcall	ImLoad		;immediate load
	IntCAP			;capture interrupt enable
	ret

;Assume new parameters:
;Input: Menu - index
	
Assume:	cpi	Menu,MnuSH
	breq	ChMd		;change mode
	cpi	Menu,MnuSY
	breq	ChMd		;change mode
	cpi	Menu,MnuD
	breq	LoadD		;change D
	cpi	Menu,MnuP
	breq	LoadP		;change P
	cpi	Menu,MnuE
	breq	LoadE		;change E
	cpi	Menu,MnuN
	breq	LoadN		;change N
	ret

;Auto sync.

syna:	rcall	LoadD		;load D to memory buffer
	rcall	LoadP		;load P to memory buffer
	rcall	ImLoad		;immediate load
SerEn:	OutToggle		;OC1 toggle
	sbrc	Mode,Mea	;meander mode, int. disable
	ret
	IntCOM			;compare interrupt enable
	ret

;Load duration to memory buffer:

LoadD:	ldi	YL,ValD
	rcall	Split		;tempC,B,A <- TN,TH,TL, tempD <- flags set
	
LdAct:	ldi	YL,Mem0
	sbrs	Mode,Loa	;check up active level
	ldi	YL,Mem1		;YL- pointer to Mem0 or Mem1
	st	Y+,tempB	;fill memory buffer
	st	Y+,tempA
	st	Y+,tempC
	st	Y+,tempD
	
;Load period to memory buffer:

LoadP:	ldi	YL,ValD
	bbrs	Mode,Mea,pme	;check up meander mode
	bbrs	Mode,Exs,Nmm	;check up external sync. mode
	rcall	LdABC		;tempC,B,A <- ValD
	ldi	YL,ValP
	rcall	LdLMN		;tempH,M,L <- ValP
	sub	tempL,tempA	;calculate inactive level duration
	sbc	tempM,tempB
	sbc	tempH,tempC
	rcall	SplitH		;tempC,B,A <- TN,TH,TL, tempD <- flags set
	rjmp	LdPas
pme:	rcall	Split	
	
LdPas:	ldi	YL,Mem0		
	sbrc	Mode,Loa	;check up active level
	ldi	YL,Mem1		;YL- pointer to Mem0 or Mem1
	st	Y+,tempB	;fill memory buffer
	st	Y+,tempA
	st	Y+,tempC
	st	Y+,tempD
Nmm:	stbr	FLAGS2,UPD	;update request
	ret
	
;Load delay to memory buffer:	
	
LoadE:	ldi	YL,ValE
	rcall	LdLMN		;tempH,M,L <- ValE
	subi	tempL,DelOff	;ValE - Delay offset
	sbci	tempM,0
	sbci	tempH,0
	rcall	SplitH		;tempC,B,A <- TN,TH,TL, tempD <- flags set
	sbrc	tempD,Sh
	stbr	tempD,ShEx	;ShEx <- Sh
	clbr	tempD,Sh
	stbr	tempD,Ex	;set Ex flag
	rjmp	LdPas
	
;Load n to NS:	
	
LoadN:	ldi	YL,ValN
	ld	temp,Y		;NS <- ValN
	sts	NS,temp
	ret

;Splits 24-bit value:
;Input: YL - pointer to value,

Split:	rcall	LdLMN		;load tempH,M,L from [Y]

;Splits 24-bit value:
;Input: tempH,M,L - value,
;       N - counter value (byte),
;       T - timer value (word)
;N.R = value / MaxW, if N=0 - short mode
;if R<(0xffff-MaxW) then R=R+MaxW, N=N-1
;R=R-1, if N=0 - short mode
;Out: TL = R(L) - tempA, TH = R(H) - tempB, TN = N    - tempC
;Input: Mode, Out:   tempD = flags set

SplitH:	ldei	MaxW		;tempF,E,D <- MaxW
	rcall	Div24		;~tempH,M,L . tempC,B,A = value / MaxW
	com	tempL
	breq	shm		;N=0
	ldi	tempH,high(0xffff-MaxW)
	ldi	tempM,low (0xffff-MaxW)
	cp	tempA,tempM
	cpc	tempB,tempH
	brsh	shm		;R>=0xffff-MaxW
	add	tempA,tempD	;R=R+MaxW
	adc	tempB,tempE
	dec	tempL		;N=N-1
shm:	ldi	tempM,1
	ldi	tempH,0	
	sub	tempA,tempM	;R=R-1
	sbc	tempB,tempH
adone:	mov	tempC,tempL

;Flags setup

	clr	tempD		;flags set init
	sbrc	Mode,Loa
	stbr	tempD,Lev	;Lev <- Loa
	tst	tempC		;check N for zero
	brne	lgmd		;N > 0, long mode
	stbr	tempD,Sh	;short mode, set Sh
lgmd:	ret
	
;Compare delta = OCR1A - TCNT1 and MinDe
;Out: C = 1 if delta < MinDe

Delta:	in	tempM,TCNT1L
	in	tempH,TCNT1H
	in	tempD,OCR1AL
	in	tempE,OCR1AH
	sub	tempD,tempM	;delta = OCR1A - TCNT1
	sbc	tempE,tempH
	subi	tempD,low (MinDe)
	sbci	tempE,high(MinDe)
	ret
	
;Immediate load interrupts parameters from memory buffer:

ImLoad:	ldi	YL,Mem0
	ld	TH0,Y+		;load TH0
	ld	TL0,Y+		;load TL0
	ld	TN0,Y+		;load TN0
	ld	FLAGS0,Y+	;load FLAGS0
	
	ld	TH1,Y+		;load TH1
	ld	TL1,Y+		;load TL1
	ld	TN1,Y+		;load TN1
	ld	FLAGS1,Y+	;load FLAGS1
	ret

;Key "EX" processing (change menu item):
	
Do_EX:	cpi	Menu,MnuDS
	brsh	Do_EN		;enter if steps
	mov	temp,Menu
	inc	temp
		
	cpi	Menu,MnuD
	brne	ex0
	sbrc	Mode,Mea
	ldi	temp,MnuSH	;Mea = 1, MnuD -> MnuSH
	bbrc	Mode,Exs,ex0
	ldi	temp,MnuSH	;Exs = 1, MnuD -> MnuSH
	sbrc	FLAGS2,EDT
	ldi	temp,MnuE	;EDT = 1, MnuD -> MnuE
	
ex0:	cpi	Menu,MnuE
	breq	ex11		;EDT = 1, MnuE -> MnuD

ex1:	cpi	Menu,MnuP
	brne	ex2
ex11:	sbrc	FLAGS2,EDT
	ldi	temp,MnuD	;EDT = 1, MnuP -> MnuD
	
ex2:	cpi	Menu,MnuSH
	brne	ex3
	sbrc	Mode,Mea	
	ldi	temp,MnuD	;Mea = 1, MnuSH -> MnuD
	
ex3:	cpi	Menu,MnuSY
	brne	ex4
	sbrs	Mode,Exs	
	ldi	temp,MnuD	;Exs = 0, MnuSY -> MnuD
	
ex4:	cpi	Menu,MnuN
	brne	ex5
	ldi	temp,MnuD	;MnuN -> MnuD

ex5:	mov	Menu,temp
ex6:	clbr	FLAGS2,EDT	;clear edited flag
ex7:	rcall	SetMd
	rjmp	EndOP		;end of processing
	
;Key "EN" processing (change menu item):

Do_EN:	cpi	Menu,MnuDS
	brlo	en1
	subi	Menu,Soff
	ldi	temp,0x00
	rcall	Sound		;return beep
	rjmp	ex7
en1:	cpi	Menu,MnuSY
	breq	ToGo
	subi	Menu,-Soff
	rjmp	ex7
	
ToGo:	bbrs	Mode,Exs,GoEn	;error if not ext. sync. mode
	rcall	ErrB		;error bell
	rjmp	EndOP
	
GoEn:	ldi	Menu,MnuGo
	IntDis			;int. disable
	lds	NSER,NS		;series counter reload
	rcall	SerEn		;enable series
	rjmp	ex7
	
;Set current edit mode:
;Input: Menu, Mode, Vals, Stps,
;Out:   Dig+0, Dig+1, Min, Max,
;       Buff <- [Vals + Menu * 3],
;       Step <- [Stps + Menu * 3],
;       UPDD = 1

SetMd:	ldi	ZL,low (StrT*2)	;string table base
	ldi	ZH,high(StrT*2)
	mov	temp,Menu
	lsl	temp		;Menu*2
	ldi	YL,Dig+0
	rcall	RdTbl		;read char 1
	st	Y+,temp
	rcall	RdTblN		;read char 2
	st	Y+,temp
	
;Load Buff from [Vals[Menu * 3]]
	
	ldi	YL,Vals
	rcall	LdLMNM		;tempH,M,L <- [Y + Menu * 3]
	ldi	YL,Buff
	rcall	StY		;[Buff] <- tempH,M,L
	
;Check for special menu cases	
	
	cpi	Menu,MnuEE
	breq	mee	
	cpi	Menu,MnuGo
	breq	mee	
	cpi	Menu,MnuSH
	breq	msh
	cpi	Menu,MnuSY
	breq	msy
	cpi	Menu,MnuDS
	brsh	stp		;Menu>=MnuDS - steps
	
;Load Step from	[Stps[Menu * 3]]
	
	ldi	YL,Stps
	rcall	LdLMNM		;tempH,M,L <- [Y + Menu * 3]
	ldi	YL,Step
	rcall	StY		;[Step] <- tempH,M,L
	
;Load Min and Max	
	
	cpi	Menu,MnuD
	breq	mnd
	cpi	Menu,MnuP
	breq	mnp
	cpi	Menu,MnuE
	breq	mne

mnn:	ldmi	MinN		;Min for n
	ldei	MaxN		;Max for n
	rjmp	stl

;Load step limits

stp:	cpi	Menu,MnuNS
	brne	stpa
	ldei	MaxNS		;Max n step
	rjmp	msm
	
stpa:	ldei	MaxS		;Max for steps
	rjmp	msm

;Special menu cases

msy:	ldei	MaxSY		;Max for SY menu			
	rjmp	msm
mee:	ldei	MaxEE		;Max for EEP and Go menu
	rjmp	msm
mne:	ldmi	MinE		;Min for E
sma:	ldei	MaxDPE		;Max for E, P and D if meander or ext.	
	rjmp	stl
msh:	ldei	MaxSH		;Max for SH menu
msm:	ldmi	1		;Step and Min	
	ldi	YL,Step
	rcall	StY		;store Step
	
;Load Min and Max from tempH,M,L and tempF,E,D
	
stl:	rcall	StMin		;store Min from tempH,M,L
	rcall	StMax		;store Max from tempF,E,D
	stbr	FLAGS2,UPDD
	ret
	
;Min and Max for D:	
	
mnd:	ldmi	MinDm		;Min for D = MinDm in meander mode
	bbrs	Mode,Mea,sma
	ldmi	MinD		;Min for D = MinD
	bbrs	Mode,Exs,sma
	ldi	YL,ValP		;Max for D = ValP - MinD
	rcall	LdDEF		;tempF,E,D <- ValP
	subi	tempD,low(MinD)
	sbci	tempE,high(MinD)
	sbci	tempF,byte3(MinD)
	rjmp	stl
	
;Min and Max for P:	
	
mnp:	ldmi	MinP		;Min for P = MinP in meander mode
	bbrs	Mode,Mea,sma
	ldi	YL,ValD		;Min for P = ValD + MinD
	rcall	LdLMN		;tempH,M,L <- ValD
	subi	tempL,low(-MinD)
	sbci	tempM,high(-MinD)
	sbci	tempH,byte3(-MinD)
	rjmp	sma
	
;Read table to temp:
;Z - base, temp - offset

RdTbl:	add	ZL,temp		;add offset (Z + temp)
	rjmp	DoRdt

;Read next byte from table to temp:
;Z - pointer

RdTblN:	adiw	ZL,1		;advance pointer
DoRdt:	clr	r0
	adc	ZH,r0
	lpm			;read next byte
	mov	temp,r0
	ret
	
;Store min limit from tempH,M,L	
	
StMin:	ldi	YL,Min

;Store 3  bytes in memory at [Y+2],[Y+1],[Y+0]
;from tempH,M,L

StY:	st	Y+,tempL
	st	Y+,tempM
	st	Y+,tempH
	ret
	
;Store max limit from tempF,E,D	
	
StMax:	ldi	YL,Max

;Store 3  bytes in memory at [Y+2],[Y+1],[Y+0]
;from tempH,M,L

StYe:	st	Y+,tempD
	st	Y+,tempE
	st	Y+,tempF
	ret
	
;Load tempH,M,L from [Y + Menu * 3]	
	
LdLMNM:	add	YL,Menu
	add	YL,Menu
	add	YL,Menu		;YL = YL + Menu * 3
	rjmp	LdLMN
	
;Display current editing value:
;Dig[3..9] <- Buff (convert to BCD) or
;Dig[3..9] <- ShpT, SynT string
;Supress leading zeros
;Add point
;LCD <- Dig[0..9]

Disp:	ldi	YL,Dig+2
	ldi	temp,BLANK
blout:	st	Y+,temp
	cpi	YL,Dig+10
	brne	blout		;blank display
	
	rcall	LdBuff		;load Buff to tempH,M,L
	
;Dig[3..8] <- strings
	
	ldi	ZL,low (ShT*2)	;shape string table base
	ldi	ZH,high(ShT*2)
	cpi	Menu,MnuSH
	breq	cops

	adiw	ZL,(SyT-ShT)*2	;sync string table base
	cpi	Menu,MnuSY
	breq	cops

	adiw	ZL,(EeT-SyT)*2	;EEP string table base
	cpi	Menu,MnuEE
	breq	cops
	
	cpi	Menu,MnuGo
	brne	notst
	adiw	ZL,(GoT-EeT)*2	;Go string table base

;Copy string (5 chars) from table to Dig[4..7]
;Z - table base, tempL - string number

cops:	mov	temp,tempL
	dec	temp
	lsl	temp		;*2
	lsl	temp		;*4
	rcall	RdTbl		;read table (byte 1)
	ldi	YL,Dig+4
cpyst:	st	Y+,temp
	rcall	RdTblN		;read next byte
	cpi	YL,Dig+8
	brne	cpyst
	rjmp	DLoad
	
;Buff[0..3] convert to BCD Dig[3..9]

notst:	ldi	ZL,Dig+10
	clr	temp
clrout: st	-Z,temp
	cpi	ZL,Dig+3
	brne	clrout		;output array clear

	ldi	tempD,24	;input bits count
	
hloop:	lsl	tempL		;input array shift left
	rol	tempM
	rol	tempH		
	ldi	YL,Dig+10
sloop:	ld	temp,-Y
	rol	temp
	subi	temp,-0x06	;temp+6, C=1
	sbrs	temp,4
	subi	temp,0x06	;temp-6, C=0
	andi	temp,0x0f
	st	Y,temp
	cpse	YL,ZL		;ZL = Dig+3
	rjmp	sloop
	dec	tempD		;YL = Dig+3
	brne	hloop
	
;Check for N or NS menu	
	
	ldi	ZL,Dig+8	
	cpi	Menu,MnuN
	breq	NoPnt
	cpi	Menu,MnuNS
	brne	Point		
NoPnt:	adiw	ZL,1
	rcall	ZSupp		;Zeros suppresion at Dig+3..Dig+9
	rjmp	DLoad
Point:	rcall	ZSupp		;Zeros suppresion at Dig+3..Dig+8
	
;Setup point	
	
	ldi	ZL,Dig+5
	cpi	YL,Dig+6	;YL points to first non-zero digit
	brlo	setpo
	ldi	ZL,Dig+8
setpo:	ld	temp,Z
	ori	temp,Pt
	st	Z,temp		;point at Dig+5 or Dig+8

;Indicate Dig[0..9]
	
DLoad:	ldi	temp,0x00	;temp <- 0x00 - SG1 address
	rcall	LCD_WA		;write address
	ldi	YL,Dig
ld1:	ld	temp,Y+		;temp <- digit

;Write symbol from temp to LCD:
;temp.7 - point

.equ	Pt	=0x80

	bst	temp,7		;T <- temp.7
	andi	temp,0x7F	;temp.7 <- 0
	ldi	ZL,low (FONT*2)	;char table base
	ldi	ZH,high(FONT*2)
	rcall	RdTbl		;r0, temp <- char
	
;Write data from r0,temp to LCD:

	rcall	LCD_WN		;write nibble from temp to LCD
	mov	temp,r0
	bld	temp,4		;bit 4 - segment H
	swap	temp
	rcall	LCD_WN		;write nibble from temp to LCD
	
	cpi	YL,Dig+10
	brne	ld1		;repeat for 10 digits
	ret	

;Zeros suppresion:
;Input: YL = Dig+3, ZL = Dig+8 or Dig+9
;Out:   [Y] 0 <- blank
;	YL points to first non-zero digit
	
ZSupp:	ldi	tempL,BLANK
zsp:	ld	temp,Y
	tst	temp
	brne	notz
	st	Y+,tempL	;blank zero
	cp	YL,ZL
	brne	zsp
notz:	ret
	
;Wait for EEPROM ready and read EEPROM:
;Input: tempE - address
;Out:   tempD - data, tempE - next address, Z=1 if tempE=EEnd

EE_Rdy: sbic	EECR,EEWE
	rjmp	EE_Rdy		;wait for EEPROM ready
	out	EEAR,tempE	;EEPROM address
	sbi	EECR,EERE	;strobe
	in	tempD,EEDR	;read EEPROM
	inc	tempE		;EEPROM address inc
	cpi	tempE,EEnd	;check address
	wdr			;watchdog restart
	ret

;***** Beeper support:

;Error bell:

ErrB:	ldi	tempE,55	;period for error bell

;Sound generation:
;Input: tempE - period

Tone:	ldi	tempF,255	;tone duration
	rjmp	Sgen	

;Sound generation:
;Input: temp - frequency

Sound:	mov	tempE,temp	;temp - frequency
	com	tempE		;temp=~temp
	andi	tempE,0x0F	;mask unused bits
	subi	tempE,-30	;add period offset
	ldi	tempF,40	;sound duration
	
;Sound generation:
;Input: tempL - period
;       tempE - duration
	
Sgen:	mov	tempA,tempE	;3
sndb:	ldi	tempD,40	  ;2 outer loop
snda:	dec	tempD		    ;1 inner loop
	brne	snda		    ;1 inner loop
	dec	tempA		  ;2 outer loop
	brne	sndb		  ;2 outer loop
	wdr			;3 watchdog restart
	sbrc	tempF,0 	;3 check tempF.0
	sbi	PORTD,SND	;3 set	 SND if tempF.0 = 1
	sbrs	tempF,0 	;3 check tempF.0
	cbi	PORTD,SND	;3 clear SND if tempF.0 = 0
	dec	tempF		;3
	brne	Sgen		;3
	ret			;tempF=1, SND = 1, beeper power off

;***** Keyboard support:

;Scan local keyboard:
;Out: temp - scancode

SCAN:	rcall	scyc		;dummy scan (load 0bXXXX1111 in HC164)
	cbi	PORTB,DATA	;DATA <- 0 at first
scyc:	ldi	temp,0b00001000	;scan 4 buttons by 5 impulses
sclk:	cbi	PORTB,CLK	;CLK <- 0
	clc			;clc inside the loop for longer delay only
	sbis	PINB,RETL	;C <- ~RETL .8uS after /CLK
	sec
	sbi	PORTB,CLK	;CLK <- 1
	sbi	PORTB,DATA	;DATA <- 1 forever
	rol	temp		;C <- temp.7..temp.0 <- C
	brcc	sclk
	ret			;temp = scancode

;***** LCD support:

;Write nibble from temp to LCD:

LCD_WN:	andi	temp,0x0F	;mask unused bits
	ori	temp,0x10	;address=1
	rjmp	LCD_W5

;Write address from temp to LCD:

LCD_WA:	andi	temp,0x0F	;mask unused bits
LCD_W5:	ldi	tempD,5		;write 5 bits to LCD
w5:	cbi	PORTB,CLK	;CLK <- 0
	cbi	PORTB,DATA	;DATA <- 0 or..
	sbrc	temp,4
	sbi	PORTB,DATA	;DATA <- 1
	rol	temp
	dec	tempD
	sbi	PORTB,CLK	;CLK <- 1
	brne	w5
	cbi	PORTB,LOAD	;LOAD <- 0
	sbi	PORTB,DATA
	sbi	PORTB,LOAD	;LOAD <- 1
	ret
		
;Font table:

FONT:	     ;FCBHADEG    FCBHADEG
	.DB 0b11101110, 0b01100000	;0, 1
	.DB 0b00101111, 0b01101101	;2, 3
	.DB 0b11100001, 0b11001101	;4, 5
	.DB 0b11001111, 0b01101000	;6, 7
	.DB 0b11101111, 0b11101101	;8, 9
	.DB 0b11101011, 0b01100111	;A, d
	.DB 0b10001111, 0b01000111	;E, o
	.DB 0b00000000, 0b00000100	;blank, _
	.DB 0b00001000,	0b11100011	;~, H
	.DB 0b10101011, 0b10000111	;P, t
	.DB 0b01000110,	0b11100101	;u, Y
	.DB 0b10000110, 0b10001010	;|_, |~
	.DB 0b11001110, 0b01000011	;G,n

;Characters codes table:

.equ	BLANK	=0x0E	;character "blank" code
.equ	i_	=0x0E	;character "blank" code
.equ	iLL	=0x0F	;character "lower -" code
.equ	iHH	=0x10	;character "upper -" code
.equ	iHL	=0x16	;character "|_" code
.equ	iLH	=0x17	;character "|~" code
.equ	iA	=0x0A	;character "A" code
.equ	iD	=0x0B	;character "d" code
.equ	iE	=0x0C	;character "E" code
.equ	iH	=0x11	;character "H" code
.equ	iO	=0x0D	;character "o" code
.equ	iP	=0x12	;character "P" code
.equ	iS	=0x05	;character "S" code
.equ	iT	=0x13	;character "t" code
.equ	iU	=0x14	;character "u" code
.equ	iY	=0x15	;character "Y" code
.equ	iG	=0x18	;character "G" code
.equ	iN	=0x19	;character "n" code

;String table:

StrT:
	.DB iD,i_,iP,i_,iS,iH,iS,iY,iE,i_,in,i_
	.DB iD,iS,iP,iS,i_,i_,i_,i_,iE,iS,in,iS

;Shape string table:

ShT:
	.DB iLL,iLH,iHL,iLH
	.DB iLL,iLH,iHL,iLL
	.DB iHH,iHL,iLH,iHH
	.DB iLL,iLL,iLL,iLL
	.DB iHH,iHH,iHH,iHH
	
;Sync string table:

SyT:
	.DB iA,iU,iT,iO
	.DB iLL,iLL,iLH,iHH
	.DB iHH,iHH,iHL,iLL
	
;EEP string table:

EeT:
	.DB iE,iE,iP,i_
	
;Go string table:

GoT:
	.DB iG,iO,i_,i_
